use super::avl::*;
use super::bounds::*;
use super::*;
use crate::func::context::*;

#[derive(Clone)]
pub struct Bound<A, T> {
    bound: T,
    bounds: Bounds<A>,
}

impl<A, T> std::ops::Deref for Bound<A, T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.bound
    }
}

pub struct BoundNode2<A, T> {
    boundsl: Bounds<A>,
    boundsr: Bounds<A>,
    bounds: Bounds<A>,
    node: T,
}

#[derive(Clone)]
pub struct BoundTrees<BT>(BT);

impl<BT> BoundTrees<BT> {
    pub fn new(bt: BT) -> Self {
        Self(bt)
    }
}

impl<'a, BT: FunctorContext<'a>> FunctorContext<'a> for BoundTrees<BT> {
    type T = BT::T;
}

pub trait BinaryTreesBindable<'a>: BinaryTrees<'a> {
    fn bounds_error<T: 'a>(&self, error: BoundsError<Self::Key>) -> BTWrap<'a, Self, T>;

    fn bounds_bind<A: 'a, B: 'a>(
        &self,
        ra: Result<A, BoundsError<Self::Key>>,
        f: impl FnOnce(A) -> BTWrap<'a, Self, B>,
    ) -> BTWrap<'a, Self, B> {
        match ra {
            Ok(a) => f(a),
            Err(e) => self.bounds_error(e),
        }
    }
}

impl<'a, BT: BinaryTreesBindable<'a>> BinaryTrees<'a> for BoundTrees<BT> {
    type Node = BoundNode2<Self::Key, BT::Node>;

    type Reference = Bound<Self::Key, BT::Reference>;

    type Tree = Bound<Self::Key, BT::Tree>;

    type Key = BT::Key;

    type Comparator = BT::Comparator;

    type _Tm = Self::T;

    fn comparator(&self) -> &Self::Comparator {
        self.0.comparator()
    }

    fn split(&self, node: &Self::Node) -> Split<'a, Self> {
        let (tl, tr, key) = self.0.split(&node.node);
        (
            Bound {
                bound: tl,
                bounds: node.boundsl.clone(),
            },
            Bound {
                bound: tr,
                bounds: node.boundsr.clone(),
            },
            key,
        )
    }

    fn resolve(&self, reference: &Self::Reference) -> BTWrap<'a, Self, Self::Node> {
        let ctx = self.0.clone();
        let bounds = reference.bounds.clone();
        Self::bind(self.0.resolve(&reference.bound), move |node| {
            let (_, _, key) = ctx.split(&node);
            ctx.bounds_bind(
                bounds.clone().split(&key, ctx.comparator()),
                |(boundsl, boundsr)| {
                    Self::pure(BoundNode2 {
                        boundsl,
                        boundsr,
                        bounds,
                        node,
                    })
                },
            )
        })
    }

    fn equal(&self, rl: &Self::Reference, rr: &Self::Reference) -> bool {
        Bounds::equal(&rl.bounds, &rr.bounds, self.comparator())
            && self.0.equal(&rl.bound, &rr.bound)
    }

    fn refer(&self, tree: &Self::Tree) -> Option<Self::Reference> {
        Some(Bound {
            bound: self.0.refer(&tree.bound)?,
            bounds: tree.bounds.clone(),
        })
    }
}

impl<'a, BT: BinaryTreesBindable<'a> + BinaryTreesTreeOf<'a>> BinaryTreesTreeOf<'a>
    for BoundTrees<BT>
{
    fn tree_of(&self, node: Self::Node) -> BTWrap<'a, Self, Self::Tree> {
        Self::fmap(self.0.tree_of(node.node), move |tree| Bound {
            bound: tree,
            bounds: node.bounds,
        })
    }
}

impl<'a, BT: BinaryTreesBindable<'a> + BinaryTreesEmpty<'a>> BinaryTreesEmpty<'a>
    for BoundTrees<BT>
{
    fn empty(&self) -> Self::Tree {
        Bound {
            bound: self.0.empty(),
            bounds: Bounds::unbound(),
        }
    }

    fn split_key_empty(
        &self,
        tree: Self::Tree,
        key: Self::Key,
    ) -> BTWrap<'a, Self, KeySplit<'a, Self>> {
        self.0.bounds_bind(
            tree.bounds.split(&key, self.comparator()),
            |(boundsl, boundsr)| {
                Self::bind(self.0.split_key_empty(tree.bound, key), |(tl, tr)| {
                    Self::pure((
                        Bound {
                            bound: tl,
                            bounds: boundsl,
                        },
                        Bound {
                            bound: tr,
                            bounds: boundsr,
                        },
                    ))
                })
            },
        )
    }
}

impl<'a, BT: BinaryTreesBindable<'a> + BinaryTreesHeight<'a>> BinaryTreesHeight<'a>
    for BoundTrees<BT>
{
    fn height(&self, tree: &Self::Tree) -> u64 {
        self.0.height(&tree.bound)
    }

    fn leaf_height_error<T: 'a>(&self, height: u64) -> BTWrap<'a, Self, T> {
        self.0.leaf_height_error(height)
    }
}

impl<'a, BT: BinaryTreesBindable<'a> + BinaryTreesTryJoin<'a>> BinaryTreesTryJoin<'a>
    for BoundTrees<BT>
{
    fn try_join(
        &self,
        tl: Self::Tree,
        key: Self::Key,
        tr: Self::Tree,
    ) -> BTWrap<'a, Self, Self::Node> {
        let (boundsl, boundsr) = (tl.bounds, tr.bounds);
        self.0.bounds_bind(
            Bounds::join(boundsl.clone(), boundsr.clone(), &key, self.comparator()),
            |bounds| {
                Self::fmap(self.0.try_join(tl.bound, key, tr.bound), |node| {
                    BoundNode2 {
                        boundsl,
                        boundsr,
                        bounds,
                        node,
                    }
                })
            },
        )
    }
}

impl<
        'a,
        BT: BinaryTreesBindable<'a>
            + BinaryTreesTreeOf<'a>
            + BinaryTreesEmpty<'a>
            + BinaryTreesHeight<'a>
            + BinaryTreesTryJoin<'a>,
    > BinaryTreesMutable<'a> for BoundTrees<BT>
{
    fn join_key(
        self,
        tl: Self::Tree,
        key: Self::Key,
        tr: Self::Tree,
    ) -> BTWrap<'a, Self, Self::Node> {
        self.join_key_balanced(tl, key, tr)
    }
}
