use std::{fmt::Display, marker::PhantomData, rc::Rc};

use crate::flow::{
    binary::{balancing::*, bound::*, *},
    comparator::*,
};
use crate::func::{context::*, *};

#[derive(Clone)]
struct Node<A> {
    l: Tree<A>,
    r: Tree<A>,
    key: A,
}

impl<A: Display> Display for Node<A> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "({} {} {})", self.l, self.key, self.r)
    }
}

struct Reference<A> {
    node: Rc<Node<A>>,
}

impl<A> Clone for Reference<A> {
    fn clone(&self) -> Self {
        Self {
            node: self.node.clone(),
        }
    }
}

impl<A: Display> Display for Reference<A> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.node)
    }
}

struct Tree<A> {
    reference: Option<Reference<A>>,
    height: u64,
}

impl<A> Clone for Tree<A> {
    fn clone(&self) -> Self {
        Self {
            reference: self.reference.clone(),
            height: self.height,
        }
    }
}

impl<A: Display> Display for Tree<A> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match &self.reference {
            Some(reference) => write!(f, "{}", reference),
            None => write!(f, "-"),
        }
    }
}

struct Trees<A> {
    _a: PhantomData<A>,
}

impl<A> Trees<A> {
    fn new() -> Self {
        Self { _a: PhantomData }
    }
}

impl<A> Clone for Trees<A> {
    fn clone(&self) -> Self {
        Self::new()
    }
}

impl<'a, A: 'a> FunctorContext<'a> for Trees<A> {
    type T = instances::solo::SoloInstance;
}

impl<'a, A: 'a + Ord + Clone> BinaryTrees<'a> for Trees<A> {
    type Node = Node<A>;

    type Reference = Reference<A>;

    type Tree = Tree<A>;

    type Key = A;

    type Comparator = DefaultComparator;

    type _Tm = Self::T;

    fn comparator(&self) -> &Self::Comparator {
        &DefaultComparator
    }

    fn split(&self, node: &Self::Node) -> Split<'a, Self> {
        (node.l.clone(), node.r.clone(), node.key.clone())
    }

    fn resolve(&self, reference: &Self::Reference) -> BTWrap<'a, Self, Self::Node> {
        reference.node.as_ref().clone()
    }

    fn equal(&self, _rhs: &Self::Reference, _lhs: &Self::Reference) -> bool {
        false
    }

    fn refer(&self, tree: &Self::Tree) -> Option<Self::Reference> {
        tree.reference.clone()
    }
}

impl<'a, A: 'a + Ord + Clone> BinaryTreesHeight<'a> for Trees<A> {
    fn height(&self, tree: &Self::Tree) -> u64 {
        tree.height
    }

    fn leaf_height_error<T: 'a>(&self, height: u64) -> BTWrap<'a, Self, T> {
        panic!("leaf height error: {height}.")
    }
}

impl<'a, A: 'a + Ord + Clone> BinaryTreesEmpty<'a> for Trees<A> {
    fn empty(&self) -> Self::Tree {
        Tree {
            reference: None,
            height: 0,
        }
    }

    fn split_key_empty(
        &self,
        _tree: Self::Tree,
        _key: Self::Key,
    ) -> BTWrap<'a, Self, KeySplit<'a, Self>> {
        (self.empty(), self.empty())
    }
}

impl<'a, A: 'a + Ord + Clone> BinaryTreesTryJoin<'a> for Trees<A> {
    fn try_join(
        &self,
        tl: Self::Tree,
        key: Self::Key,
        tr: Self::Tree,
    ) -> BTWrap<'a, Self, Self::Node> {
        Node { l: tl, r: tr, key }
    }
}

impl<'a, A: 'a + Ord + Clone> BinaryTreesUnbalanced<'a> for Trees<A> {
    fn tree_of_with_height(&self, node: Self::Node, height: u64) -> BTWrap<'a, Self, Self::Tree> {
        Tree {
            reference: Some(Reference {
                node: Rc::new(node),
            }),
            height,
        }
    }

    fn balancing_error<T: 'a>(&self, _error: BalancingError) -> BTWrap<'a, Self, T> {
        panic!("balancing error.")
    }

    fn node_heights(&self, node: &Self::Node) -> (u64, u64) {
        (node.l.height, node.r.height)
    }
}

impl<'a, A: 'a + Ord + Clone> BinaryTreesBindable<'a> for Trees<A> {
    fn bounds_error<T: 'a>(&self, _error: bounds::BoundsError<Self::Key>) -> BTWrap<'a, Self, T> {
        panic!("bounds violated.");
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    impl<A> Node<A> {
        fn balanced(&self) -> bool {
            let (hl, hr) = (self.l.height, self.r.height);
            std::cmp::max(hl, hr) - std::cmp::min(hl, hr) < 2
                && self.l.balanced()
                && self.r.balanced()
        }
    }

    impl<A> Reference<A> {
        fn balanced(&self) -> bool {
            self.node.balanced()
        }
    }

    impl<A> Tree<A> {
        fn balanced(&self) -> bool {
            match &self.reference {
                Some(reference) => reference.balanced(),
                None => true,
            }
        }
    }

    #[test]
    fn test() {
        let trees = BalancedTrees::new(Trees::new());
        let mut tree = trees.empty();
        for i in [
            8, 3, 10, 17, 0, 13, 6, 1, 11, 5, 4, 7, 18, 14, 15, 9, 2, 19, 16, 12,
        ] {
            tree = trees.clone().add_tree(tree, i);
            assert!(tree.balanced());
            // println!("{} {}", tree.height, tree);
        }
        for i in [
            2, 9, 4, 7, 8, 10, 17, 1, 13, 15, 18, 12, 5, 0, 3, 6, 16, 19, 14, 11,
        ] {
            tree = trees.clone().remove(tree, i);
            assert!(tree.balanced());
            // println!("{} {}", tree.height, tree);
        }
        // assert!(false);
    }

    #[test]
    fn test_with_bounds() {
        let trees = BoundTrees::new(BalancedTrees::new(Trees::new()));
        let mut tree = trees.empty();
        for i in [
            8, 3, 10, 17, 0, 13, 6, 1, 11, 5, 4, 7, 18, 14, 15, 9, 2, 19, 16, 12,
        ] {
            tree = trees.clone().add_tree(tree, i);
            assert!(tree.balanced());
            // println!("{} {}", tree.height, tree);
        }
        for i in [
            2, 9, 4, 7, 8, 10, 17, 1, 13, 15, 18, 12, 5, 0, 3, 6, 16, 19, 14, 11,
        ] {
            tree = trees.clone().remove(tree, i);
            assert!(tree.balanced());
            // println!("{} {}", tree.height, tree);
        }
        // assert!(false);
    }
}
